(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * FWTesting.thy ---
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************.***************************************************************)
(* $Id: Section5Example.thy 9105 2011-01-25 22:03:54Z makarius $ *)

theory Section5Example
imports FWTesting
begin

text{* This is the example which serves as running example in the ICST
2010 paper. *}

declare [[testgen_profiling]]

definition internet::"IntegerPort net" where "internet \<equiv> {{(a,b). a = 7}}"

definition dmz::"IntegerPort net" where "dmz \<equiv> {{(a,b). a= 6}}"

definition intranet::"IntegerPort net" where "intranet \<equiv> {{(a,b). a= 5}}"

definition Policy :: "(IntegerPort net, port) Combinators" where 
"Policy \<equiv> DenyAll \<oplus>
          AllowPortFromTo intranet internet 8 \<oplus>
          AllowPortFromTo intranet dmz 443 \<oplus>
          AllowPortFromTo dmz intranet 25 \<oplus>
          AllowPortFromTo intranet dmz 25 \<oplus>
          AllowPortFromTo internet dmz 80 \<oplus>
          AllowPortFromTo internet dmz 25"


lemma sets_distinct1: "(n::int) \<noteq> m \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
by auto

lemma sets_distinct2: "(m::int) \<noteq> n \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
by auto

lemma sets_distinct5: "(n::int) < m \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
by auto

lemma sets_distinct6: "(m::int) < n \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
by auto

lemma nets_different: "internet \<noteq> intranet \<and> internet \<noteq> dmz \<and> intranet \<noteq> dmz \<and>
                       intranet \<noteq> internet \<and> dmz \<noteq> internet \<and> dmz \<noteq> intranet"
apply (simp add: internet_def intranet_def dmz_def )
apply (simp add: sets_distinct1 sets_distinct2 sets_distinct5 sets_distinct6)
done


 text{*
Two auxiliary lemmas
*}
 
lemma aux: "\<lbrakk>x \<noteq> a; y\<noteq>b; (x \<noteq> y \<and> x \<noteq> b) \<or>  (a \<noteq> b \<and> a \<noteq> y)\<rbrakk> \<Longrightarrow> {x,a} \<noteq> {y,b}"
  apply auto
done


lemma aux2: "{a,b} = {b,a}"
  apply auto
done



lemma noMT: "\<forall> x \<in> set (policy2list Policy). dom (C x) \<noteq> {}" 
apply (simp add: Policy_def)
apply (simp add: if_splits PLemmas internet_def intranet_def dmz_def
  IntegerPort.src_port allow_all_def deny_all_def)
done



lemma ANDex: "allNetsDistinct (policy2list Policy)"
apply (simp add: Policy_def allNetsDistinct_def nets_different)
apply (auto simp: PLemmas internet_def intranet_def dmz_def  netsDistinct_def
  IntegerPort.src_port)
done


lemma count_the_rules: 
  "(int (length(policy2list (list2policy(normalize Policy)))) = post) \<and>
   (int(length (policy2list Policy)) = pre) \<and>
   (int (length((normalize Policy))) = Partitions)"
apply (insert nets_different noMT)
apply (simp add:  normalize_def Policy_def bothNets_def aux aux2 Nets_List_def,
        thin_tac "?X",thin_tac "?S")
oops



lemma all_in_list: "all_in_list (policy2list Policy) (Nets_List Policy)"
apply (simp add: Policy_def)
apply (unfold Nets_List_def)
apply (unfold bothNets_def)
apply (rule conjI)
apply (insert nets_different)
apply simp
apply (simp add: aux2)
done



lemma C_Policy_compile: " 
C (list2policy (normalize Policy)) =  C Policy"
apply (unfold normalize_def)
apply (rule C_eq_compile)
apply (simp add: Policy_def)
apply (rule all_in_list)
apply (rule ANDex)
done




test_spec "makeFUT FUT Policy x 0"
apply (simp add: makeFUT_def)
apply (rule_tac x=x in spec, rule allI, simp only: split_tupled_all, rule impI)
apply (insert nets_different noMT)
apply (simp add:  normalize_def Policy_def bothNets_def aux aux2 Nets_List_def)
apply (rotate_tac -1, thin_tac "?x")
apply (rotate_tac -1, thin_tac "?x")
apply (simp add: packet_Nets_def)
apply (simp add: PLemmas internet_def intranet_def dmz_def    IntegerPort.src_port)
apply (gen_test_cases "FUT")
apply simp_all
store_test_thm "foo0"
gen_test_data "foo0"

thm foo0.test_data



test_spec "makeFUT FUT Policy x 1"
apply (simp add: makeFUT_def)
apply (rule_tac x=x in spec, rule allI, simp only: split_tupled_all, rule impI)

apply (insert nets_different noMT)
apply (simp add:  normalize_def Policy_def bothNets_def aux aux2 Nets_List_def)
apply (rotate_tac -1, thin_tac "?x")
apply (rotate_tac -1, thin_tac "?x")
apply (simp add: packet_Nets_def)
apply (simp add: PLemmas internet_def intranet_def dmz_def    IntegerPort.src_port)
apply (gen_test_cases "FUT")
apply simp_all
store_test_thm "foo1"
gen_test_data "foo1"

thm foo1.test_data


test_spec "makeFUT FUT Policy x 2"
apply (simp add: makeFUT_def)
apply (rule_tac x=x in spec, rule allI, simp only: split_tupled_all, rule impI)

apply (insert nets_different noMT)
apply (simp add:  normalize_def Policy_def bothNets_def aux aux2 Nets_List_def)
apply (rotate_tac -1, thin_tac "?x")
apply (rotate_tac -1, thin_tac "?x")
apply (simp add: packet_Nets_def)
apply (simp add: PLemmas internet_def intranet_def dmz_def    IntegerPort.src_port)
apply (gen_test_cases "FUT")
apply simp_all
store_test_thm "foo2"

gen_test_data "foo2"

thm foo0.test_data
thm foo1.test_data
thm foo2.test_data


test_spec "FUT x = C (Policy) x"
apply (rule_tac x = x in spec, rule allI)
apply (simp only: split_tupled_all)
apply (simp add: Policy_def)
apply (simp add: C.simps)
apply (simp add: PLemmas internet_def intranet_def dmz_def    IntegerPort.src_port)
apply (gen_test_cases "FUT")
apply simp_all
store_test_thm "foo"
gen_test_data "foo"

thm foo.test_data

write_clocks "time_stats_Section5Example"

end
