(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * FWTesting.thy ---
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************.***************************************************************)
(* $Id: Section5Example.thy 9105 2011-01-25 22:03:54Z makarius $ *)

theory NATNormalisationTest
imports FWTesting
begin


text{* This theory contains the definitions and theorems which are
used in the case studies for the normalisation procedure. *}


fun dupl :: "(IntegerPort net,port)Combinators list \<Rightarrow> int" where 
 "dupl (x#xs) = (if x \<in> set xs then ((dupl xs) + (1::int)) else (dupl xs))"
|"dupl [] = 0" 



definition one :: "(IntegerPort net)" where
"one = {{(a,b). (a = 1)}}"
definition two :: "(IntegerPort net)" where
"two = {{(a,b). (a = 2)}}"
definition three :: "(IntegerPort net)" where
"three = {{(a,b). (a = 3)}}"
definition four :: "(IntegerPort net)" where
"four = {{(a,b). (a = 4)}}"
definition five :: "(IntegerPort net)" where
"five = {{(a,b). (a = 5)}}"
definition six :: "(IntegerPort net)" where
"six = {{(a,b). (a = 6)}}"
definition seven :: "(IntegerPort net)" where
"seven = {{(a,b). (a = 7)}}"
definition eight :: "(IntegerPort net)" where
"eight = {{(a,b). (a = 8)}}"
definition nine :: "(IntegerPort net)" where
"nine = {{(a,b). (a = 9)}}"
definition ten :: "(IntegerPort net)" where
"ten = {{(a,b). (a = 10)}}"
definition eleven :: "(IntegerPort net)" where
"eleven = {{(a,b). (a = 11)}}"
definition twelve :: "(IntegerPort net)" where
"twelve = {{(a,b). (a = 12)}}"
definition thirteen :: "(IntegerPort net)" where
"thirteen = {{(a,b). (a = 13)}}"
definition fourteen :: "(IntegerPort net)" where
"fourteen = {{(a,b). (a = 14)}}"
definition fifteen :: "(IntegerPort net)" where
"fifteen = {{(a,b). (a = 15)}}"
definition sixteen :: "(IntegerPort net)" where
"sixteen = {{(a,b). (a = 16)}}"
definition seventeen :: "(IntegerPort net)" where
"seventeen = {{(a,b). (a = 17)}}"
definition eighteen :: "(IntegerPort net)" where
"eighteen = {{(a,b). (a = 18)}}"
definition nineteen :: "(IntegerPort net)" where
"nineteen = {{(a,b). (a = 19)}}"
definition twenty :: "(IntegerPort net)" where
"twenty = {{(a,b). (a = 20)}}"
definition twentyone :: "(IntegerPort net)" where
"twentyone = {{(a,b). (a = 21)}}"
definition twentytwo :: "(IntegerPort net)" where
"twentytwo = {{(a,b). (a = 22)}}"
definition twentythree :: "(IntegerPort net)" where
"twentythree = {{(a,b). (a = 23)}}"
definition twentyfour :: "(IntegerPort net)" where
"twentyfour = {{(a,b). (a = 24)}}"
definition twentyfive :: "(IntegerPort net)" where
"twentyfive = {{(a,b). (a = 25)}}"
definition twentysix :: "(IntegerPort net)" where
"twentysix = {{(a,b). (a = 26)}}"
definition twentyseven :: "(IntegerPort net)" where
"twentyseven = {{(a,b). (a = 27)}}"
definition twentyeight :: "(IntegerPort net)" where
"twentyeight = {{(a,b). (a = 28)}}"
definition twentynine :: "(IntegerPort net)" where
"twentynine = {{(a,b). (a = 29)}}"
definition thirty :: "(IntegerPort net)" where
"thirty = {{(a,b). (a = 30)}}"
definition thirtyone :: "(IntegerPort net)" where
"thirtyone = {{(a,b). (a = 31)}}"





lemmas UnfoldPolicy = one_def two_def three_def four_def five_def
six_def seven_def eight_def nine_def ten_def eleven_def twelve_def
thirteen_def fourteen_def fifteen_def sixteen_def seventeen_def
eighteen_def nineteen_def twenty_def twentyone_def twentytwo_def
twentythree_def twentyfour_def twentyfive_def twentysix_def
twentyseven_def twentyeight_def twentynine_def thirty_def
thirtyone_def


lemma sets_distinct1: "n \<noteq> m \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
apply auto
done

lemma sets_distinct2: "m \<noteq> n \<Longrightarrow> {(a,b). a = n} \<noteq> {(a,b). a = m}"
apply auto
done


lemma nets_different: 
"one \<noteq> two \<and> one \<noteq> three
\<and> one \<noteq> four \<and> one \<noteq> five \<and> one \<noteq>
six \<and> one \<noteq> seven \<and> one \<noteq> eight \<and> one
\<noteq> nine \<and> one \<noteq> ten \<and> one \<noteq> eleven
\<and> one \<noteq> twelve \<and> one \<noteq> thirteen \<and> one \<noteq>
fourteen \<and> one \<noteq> fifteen \<and> one \<noteq> sixteen \<and> one
\<noteq> seventeen \<and> one \<noteq> eighteen \<and> one \<noteq>
nineteen \<and> one \<noteq> twenty \<and> one \<noteq> twentyone
\<and> one \<noteq> twentytwo \<and> one \<noteq> twentythree \<and>
one \<noteq> twentyfour \<and> one \<noteq> twentyfive \<and> one
\<noteq> twentysix \<and> one \<noteq> twentyseven \<and> one \<noteq>
twentyeight \<and> one \<noteq> twentynine \<and> one \<noteq> thirty
\<and> one \<noteq> thirtyone \<and> two \<noteq> one \<and> two
\<noteq> three \<and> two \<noteq> four \<and> two \<noteq> five
\<and> two \<noteq> six \<and> two \<noteq> seven \<and> two \<noteq>
eight \<and> two \<noteq> nine \<and> two \<noteq> ten \<and> two
\<noteq> eleven \<and> two \<noteq> twelve \<and> two \<noteq> thirteen
\<and> two \<noteq> fourteen \<and> two \<noteq> fifteen \<and> two
\<noteq> sixteen \<and> two \<noteq> seventeen \<and> two \<noteq>
eighteen \<and> two \<noteq> nineteen \<and> two \<noteq> twenty
\<and> two \<noteq> twentyone \<and> two \<noteq> twentytwo \<and> two
\<noteq> twentythree \<and> two \<noteq> twentyfour \<and> two
\<noteq> twentyfive \<and> two \<noteq> twentysix \<and> two \<noteq>
twentyseven \<and> two \<noteq> twentyeight \<and> two \<noteq>
twentynine \<and> two \<noteq> thirty \<and> two \<noteq> thirtyone
\<and> three \<noteq> one \<and> three \<noteq> two \<and> three
\<noteq> four \<and> three \<noteq> five \<and> three \<noteq> six
\<and> three \<noteq> seven \<and> three \<noteq> eight \<and> three
\<noteq> nine \<and> three \<noteq> ten \<and> three \<noteq> eleven
\<and> three \<noteq> twelve \<and> three \<noteq> thirteen \<and> three
\<noteq> fourteen \<and> three \<noteq> fifteen \<and> three \<noteq>
sixteen \<and> three \<noteq> seventeen \<and> three \<noteq> eighteen
\<and> three \<noteq> nineteen \<and> three \<noteq> twenty \<and>
three \<noteq> twentyone \<and> three \<noteq> twentytwo \<and> three
\<noteq> twentythree \<and> three \<noteq> twentyfour \<and> three
\<noteq> twentyfive \<and> three \<noteq> twentysix \<and> three
\<noteq> twentyseven \<and> three \<noteq> twentyeight \<and> three
\<noteq> twentynine \<and> three \<noteq> thirty \<and> three \<noteq>
thirtyone \<and> four \<noteq> one \<and> four \<noteq> two \<and>
four \<noteq> three \<and> four \<noteq> five \<and> four \<noteq> six
\<and> four \<noteq> seven \<and> four \<noteq> eight \<and> four
\<noteq> nine \<and> four \<noteq> ten \<and> four \<noteq> eleven
\<and> four \<noteq> twelve \<and> four \<noteq> thirteen \<and> four
\<noteq> fourteen \<and> four \<noteq> fifteen \<and> four \<noteq> sixteen
\<and> four \<noteq> seventeen \<and> four \<noteq> eighteen \<and>
four \<noteq> nineteen \<and> four \<noteq> twenty \<and> four
\<noteq> twentyone \<and> four \<noteq> twentytwo \<and> four \<noteq>
twentythree \<and> four \<noteq> twentyfour \<and> four \<noteq>
twentyfive \<and> four \<noteq> twentysix \<and> four \<noteq>
twentyseven \<and> four \<noteq> twentyeight \<and> four \<noteq>
twentynine \<and> four \<noteq> thirty \<and> four \<noteq> thirtyone
\<and> five \<noteq> one \<and> five \<noteq> two \<and> five \<noteq>
three \<and> five \<noteq> four \<and> five \<noteq> six \<and> five
\<noteq> seven \<and> five \<noteq> eight \<and> five \<noteq> nine
\<and> five \<noteq> ten \<and> five \<noteq> eleven \<and> five
\<noteq> twelve \<and> five \<noteq> thirteen \<and> five \<noteq> fourteen
\<and> five \<noteq> fifteen \<and> five \<noteq> sixteen \<and> five
\<noteq> seventeen \<and> five \<noteq> eighteen \<and> five \<noteq>
nineteen \<and> five \<noteq> twenty \<and> five \<noteq> twentyone
\<and> five \<noteq> twentytwo \<and> five \<noteq> twentythree \<and>
five \<noteq> twentyfour \<and> five \<noteq> twentyfive \<and> five
\<noteq> twentysix \<and> five \<noteq> twentyseven \<and> five
\<noteq> twentyeight \<and> five \<noteq> twentynine \<and> five
\<noteq> thirty \<and> five \<noteq> thirtyone \<and> six \<noteq> one
\<and> six \<noteq> two \<and> six \<noteq> three \<and> six \<noteq>
four \<and> six \<noteq> five \<and> six \<noteq> seven \<and> six
\<noteq> eight \<and> six \<noteq> nine \<and> six \<noteq> ten \<and>
six \<noteq> eleven \<and> six \<noteq> twelve \<and> six \<noteq> thirteen
\<and> six \<noteq> fourteen \<and> six \<noteq> fifteen \<and> six
\<noteq> sixteen \<and> six \<noteq> seventeen \<and> six \<noteq>
eighteen \<and> six \<noteq> nineteen \<and> six \<noteq> twenty
\<and> six \<noteq> twentyone \<and> six \<noteq> twentytwo \<and> six
\<noteq> twentythree \<and> six \<noteq> twentyfour \<and> six
\<noteq> twentyfive \<and> six \<noteq> twentysix \<and> six \<noteq>
twentyseven \<and> six \<noteq> twentyeight \<and> six \<noteq>
twentynine \<and> six \<noteq> thirty \<and> six \<noteq> thirtyone
\<and> seven \<noteq> one \<and> seven \<noteq> two \<and> seven
\<noteq> three \<and> seven \<noteq> four \<and> seven \<noteq> five
\<and> seven \<noteq> six \<and> seven \<noteq> eight \<and> seven
\<noteq> nine \<and> seven \<noteq> ten \<and> seven \<noteq> eleven
\<and> seven \<noteq> twelve \<and> seven \<noteq> thirteen \<and> seven
\<noteq> fourteen \<and> seven \<noteq> fifteen \<and> seven \<noteq>
sixteen \<and> seven \<noteq> seventeen \<and> seven \<noteq> eighteen
\<and> seven \<noteq> nineteen \<and> seven \<noteq> twenty \<and>
seven \<noteq> twentyone \<and> seven \<noteq> twentytwo \<and> seven
\<noteq> twentythree \<and> seven \<noteq> twentyfour \<and> seven
\<noteq> twentyfive \<and> seven \<noteq> twentysix \<and> seven
\<noteq> twentyseven \<and> seven \<noteq> twentyeight \<and> seven
\<noteq> twentynine \<and> seven \<noteq> thirty \<and> seven \<noteq>
thirtyone \<and> eight \<noteq> one \<and> eight \<noteq> two \<and>
eight \<noteq> three \<and> eight \<noteq> four \<and> eight \<noteq>
five \<and> eight \<noteq> six \<and> eight \<noteq> seven \<and>
eight \<noteq> nine \<and> eight \<noteq> ten \<and> eight \<noteq>
eleven \<and> eight \<noteq> twelve \<and> eight \<noteq> thirteen \<and>
eight \<noteq> fourteen \<and> eight \<noteq> fifteen \<and> eight \<noteq>
sixteen \<and> eight \<noteq> seventeen \<and> eight \<noteq> eighteen
\<and> eight \<noteq> nineteen \<and> eight \<noteq> twenty \<and>
eight \<noteq> twentyone \<and> eight \<noteq> twentytwo \<and> eight
\<noteq> twentythree \<and> eight \<noteq> twentyfour \<and> eight
\<noteq> twentyfive \<and> eight \<noteq> twentysix \<and> eight
\<noteq> twentyseven \<and> eight \<noteq> twentyeight \<and> eight
\<noteq> twentynine \<and> eight \<noteq> thirty \<and> eight \<noteq>
thirtyone \<and> nine \<noteq> one \<and> nine \<noteq> two \<and>
nine \<noteq> three \<and> nine \<noteq> four \<and> nine \<noteq>
five \<and> nine \<noteq> six \<and> nine \<noteq> seven \<and> nine
\<noteq> eight \<and> nine \<noteq> ten \<and> nine \<noteq> eleven
\<and> nine \<noteq> twelve \<and> nine \<noteq> thirteen \<and> nine
\<noteq> fourteen \<and> nine \<noteq> fifteen \<and> nine \<noteq> sixteen
\<and> nine \<noteq> seventeen \<and> nine \<noteq> eighteen \<and>
nine \<noteq> nineteen \<and> nine \<noteq> twenty \<and> nine
\<noteq> twentyone \<and> nine \<noteq> twentytwo \<and> nine \<noteq>
twentythree \<and> nine \<noteq> twentyfour \<and> nine \<noteq>
twentyfive \<and> nine \<noteq> twentysix \<and> nine \<noteq>
twentyseven \<and> nine \<noteq> twentyeight \<and> nine \<noteq>
twentynine \<and> nine \<noteq> thirty \<and> nine \<noteq> thirtyone
\<and> ten \<noteq> one \<and> ten \<noteq> two \<and> ten \<noteq>
three \<and> ten \<noteq> four \<and> ten \<noteq> five \<and> ten
\<noteq> six \<and> ten \<noteq> seven \<and> ten \<noteq> eight
\<and> ten \<noteq> nine \<and> ten \<noteq> eleven \<and> ten
\<noteq> twelve \<and> ten \<noteq> thirteen \<and> ten \<noteq> fourteen
\<and> ten \<noteq> fifteen \<and> ten \<noteq> sixteen \<and> ten \<noteq>
seventeen \<and> ten \<noteq> eighteen \<and> ten \<noteq> nineteen
\<and> ten \<noteq> twenty \<and> ten \<noteq> twentyone \<and> ten
\<noteq> twentytwo \<and> ten \<noteq> twentythree \<and> ten \<noteq>
twentyfour \<and> ten \<noteq> twentyfive \<and> ten \<noteq>
twentysix \<and> ten \<noteq> twentyseven \<and> ten \<noteq>
twentyeight \<and> ten \<noteq> twentynine \<and> ten \<noteq> thirty
\<and> ten \<noteq> thirtyone \<and> eleven \<noteq> one \<and> eleven
\<noteq> two \<and> eleven \<noteq> three \<and> eleven \<noteq> four
\<and> eleven \<noteq> five \<and> eleven \<noteq> six \<and> eleven
\<noteq> seven \<and> eleven \<noteq> eight \<and> eleven \<noteq>
nine \<and> eleven \<noteq> ten \<and> eleven \<noteq> twelve \<and>
eleven \<noteq> thirteen \<and> eleven \<noteq> fourteen \<and> eleven
\<noteq> fifteen \<and> eleven \<noteq> sixteen \<and> eleven \<noteq>
seventeen \<and> eleven \<noteq> eighteen \<and> eleven \<noteq>
nineteen \<and> eleven \<noteq> twenty \<and> eleven \<noteq>
twentyone \<and> eleven \<noteq> twentytwo \<and> eleven \<noteq>
twentythree \<and> eleven \<noteq> twentyfour \<and> eleven \<noteq>
twentyfive \<and> eleven \<noteq> twentysix \<and> eleven \<noteq>
twentyseven \<and> eleven \<noteq> twentyeight \<and> eleven \<noteq>
twentynine \<and> eleven \<noteq> thirty \<and> eleven \<noteq>
thirtyone \<and> twelve \<noteq> one \<and> twelve \<noteq> two \<and>
twelve \<noteq> three \<and> twelve \<noteq> four \<and> twelve
\<noteq> five \<and> twelve \<noteq> six \<and> twelve \<noteq> seven
\<and> twelve \<noteq> eight \<and> twelve \<noteq> nine \<and> twelve
\<noteq> ten \<and> twelve \<noteq> eleven \<and> twelve \<noteq> thirteen
\<and> twelve \<noteq> fourteen \<and> twelve \<noteq> fifteen \<and>
twelve \<noteq> sixteen \<and> twelve \<noteq> seventeen \<and> twelve
\<noteq> eighteen \<and> twelve \<noteq> nineteen \<and> twelve
\<noteq> twenty \<and> twelve \<noteq> twentyone \<and> twelve
\<noteq> twentytwo \<and> twelve \<noteq> twentythree \<and> twelve
\<noteq> twentyfour \<and> twelve \<noteq> twentyfive \<and> twelve
\<noteq> twentysix \<and> twelve \<noteq> twentyseven \<and> twelve
\<noteq> twentyeight \<and> twelve \<noteq> twentynine \<and> twelve
\<noteq> thirty \<and> twelve \<noteq> thirtyone \<and> thirteen \<noteq>
one \<and> thirteen \<noteq> two \<and> thirteen \<noteq> three \<and> thirteen \<noteq>
four \<and> thirteen \<noteq> five \<and> thirteen \<noteq> six \<and> thirteen \<noteq>
seven \<and> thirteen \<noteq> eight \<and> thirteen \<noteq> nine \<and> thirteen
\<noteq> ten \<and> thirteen \<noteq> eleven \<and> thirteen \<noteq> twelve
\<and> thirteen \<noteq> fourteen \<and> thirteen \<noteq> fifteen \<and> thirteen \<noteq>
sixteen \<and> thirteen \<noteq> seventeen \<and> thirteen \<noteq> eighteen
\<and> thirteen \<noteq> nineteen \<and> thirteen \<noteq> twenty \<and> thirteen
\<noteq> twentyone \<and> thirteen \<noteq> twentytwo \<and> thirteen \<noteq>
twentythree \<and> thirteen \<noteq> twentyfour \<and> thirteen \<noteq>
twentyfive \<and> thirteen \<noteq> twentysix \<and> thirteen \<noteq> twentyseven
\<and> thirteen \<noteq> twentyeight \<and> thirteen \<noteq> twentynine \<and> thirteen
\<noteq> thirty \<and> thirteen \<noteq> thirtyone \<and> fourteen \<noteq>
one \<and> fourteen \<noteq> two \<and> fourteen \<noteq> three \<and>
fourteen \<noteq> four \<and> fourteen \<noteq> five \<and> fourteen
\<noteq> six \<and> fourteen \<noteq> seven \<and> fourteen \<noteq>
eight \<and> fourteen \<noteq> nine \<and> fourteen \<noteq> ten
\<and> fourteen \<noteq> eleven \<and> fourteen \<noteq> twelve \<and>
fourteen \<noteq> thirteen \<and> fourteen \<noteq> fifteen \<and> fourteen
\<noteq> sixteen \<and> fourteen \<noteq> seventeen \<and> fourteen
\<noteq> eighteen \<and> fourteen \<noteq> nineteen \<and> fourteen
\<noteq> twenty \<and> fourteen \<noteq> twentyone \<and> fourteen
\<noteq> twentytwo \<and> fourteen \<noteq> twentythree \<and>
fourteen \<noteq> twentyfour \<and> fourteen \<noteq> twentyfive
\<and> fourteen \<noteq> twentysix \<and> fourteen \<noteq>
twentyseven \<and> fourteen \<noteq> twentyeight \<and> fourteen
\<noteq> twentynine \<and> fourteen \<noteq> thirty \<and> fourteen
\<noteq> thirtyone \<and> fifteen \<noteq> one \<and> fifteen \<noteq> two
\<and> fifteen \<noteq> three \<and> fifteen \<noteq> four \<and> fifteen \<noteq>
five \<and> fifteen \<noteq> six \<and> fifteen \<noteq> seven \<and> fifteen
\<noteq> eight \<and> fifteen \<noteq> nine \<and> fifteen \<noteq> ten \<and>
fifteen \<noteq> eleven \<and> fifteen \<noteq> twelve \<and> fifteen \<noteq> thirteen
\<and> fifteen \<noteq> fourteen \<and> fifteen \<noteq> sixteen \<and> fifteen
\<noteq> seventeen \<and> fifteen \<noteq> eighteen \<and> fifteen \<noteq>
nineteen \<and> fifteen \<noteq> twenty \<and> fifteen \<noteq> twentyone \<and>
fifteen \<noteq> twentytwo \<and> fifteen \<noteq> twentythree \<and> fifteen
\<noteq> twentyfour \<and> fifteen \<noteq> twentyfive \<and> fifteen \<noteq>
twentysix \<and> fifteen \<noteq> twentyseven \<and> fifteen \<noteq>
twentyeight \<and> fifteen \<noteq> twentynine \<and> fifteen \<noteq> thirty
\<and> fifteen \<noteq> thirtyone \<and> sixteen \<noteq> one \<and>
sixteen \<noteq> two \<and> sixteen \<noteq> three \<and> sixteen
\<noteq> four \<and> sixteen \<noteq> five \<and> sixteen \<noteq> six
\<and> sixteen \<noteq> seven \<and> sixteen \<noteq> eight \<and>
sixteen \<noteq> nine \<and> sixteen \<noteq> ten \<and> sixteen
\<noteq> eleven \<and> sixteen \<noteq> twelve \<and> sixteen \<noteq>
thirteen \<and> sixteen \<noteq> fourteen \<and> sixteen \<noteq> fifteen \<and>
sixteen \<noteq> seventeen \<and> sixteen \<noteq> eighteen \<and>
sixteen \<noteq> nineteen \<and> sixteen \<noteq> twenty \<and>
sixteen \<noteq> twentyone \<and> sixteen \<noteq> twentytwo \<and>
sixteen \<noteq> twentythree \<and> sixteen \<noteq> twentyfour \<and>
sixteen \<noteq> twentyfive \<and> sixteen \<noteq> twentysix \<and>
sixteen \<noteq> twentyseven \<and> sixteen \<noteq> twentyeight
\<and> sixteen \<noteq> twentynine \<and> sixteen \<noteq> thirty
\<and> sixteen \<noteq> thirtyone \<and> seventeen \<noteq> one \<and>
seventeen \<noteq> two \<and> seventeen \<noteq> three \<and>
seventeen \<noteq> four \<and> seventeen \<noteq> five \<and>
seventeen \<noteq> six \<and> seventeen \<noteq> seven \<and>
seventeen \<noteq> eight \<and> seventeen \<noteq> nine \<and>
seventeen \<noteq> ten \<and> seventeen \<noteq> eleven \<and>
seventeen \<noteq> twelve \<and> seventeen \<noteq> thirteen \<and>
seventeen \<noteq> fourteen \<and> seventeen \<noteq> fifteen \<and>
seventeen \<noteq> sixteen \<and> seventeen \<noteq> eighteen \<and>
seventeen \<noteq> nineteen \<and> seventeen \<noteq> twenty \<and>
seventeen \<noteq> twentyone \<and> seventeen \<noteq> twentytwo
\<and> seventeen \<noteq> twentythree \<and> seventeen \<noteq>
twentyfour \<and> seventeen \<noteq> twentyfive \<and> seventeen
\<noteq> twentysix \<and> seventeen \<noteq> twentyseven \<and>
seventeen \<noteq> twentyeight \<and> seventeen \<noteq> twentynine
\<and> seventeen \<noteq> thirty \<and> seventeen \<noteq> thirtyone
\<and> eighteen \<noteq> one \<and> eighteen \<noteq> two \<and>
eighteen \<noteq> three \<and> eighteen \<noteq> four \<and> eighteen
\<noteq> five \<and> eighteen \<noteq> six \<and> eighteen \<noteq>
seven \<and> eighteen \<noteq> eight \<and> eighteen \<noteq> nine
\<and> eighteen \<noteq> ten \<and> eighteen \<noteq> eleven \<and>
eighteen \<noteq> twelve \<and> eighteen \<noteq> thirteen \<and> eighteen
\<noteq> fourteen \<and> eighteen \<noteq> fifteen \<and> eighteen \<noteq>
sixteen \<and> eighteen \<noteq> seventeen \<and> eighteen \<noteq>
nineteen \<and> eighteen \<noteq> twenty \<and> eighteen \<noteq>
twentyone \<and> eighteen \<noteq> twentytwo \<and> eighteen \<noteq>
twentythree \<and> eighteen \<noteq> twentyfour \<and> eighteen
\<noteq> twentyfive \<and> eighteen \<noteq> twentysix \<and> eighteen
\<noteq> twentyseven \<and> eighteen \<noteq> twentyeight \<and>
eighteen \<noteq> twentynine \<and> eighteen \<noteq> thirty \<and>
eighteen \<noteq> thirtyone \<and> nineteen \<noteq> one \<and>
nineteen \<noteq> two \<and> nineteen \<noteq> three \<and> nineteen
\<noteq> four \<and> nineteen \<noteq> five \<and> nineteen \<noteq>
six \<and> nineteen \<noteq> seven \<and> nineteen \<noteq> eight
\<and> nineteen \<noteq> nine \<and> nineteen \<noteq> ten \<and>
nineteen \<noteq> eleven \<and> nineteen \<noteq> twelve \<and>
nineteen \<noteq> thirteen \<and> nineteen \<noteq> fourteen \<and> nineteen
\<noteq> fifteen \<and> nineteen \<noteq> sixteen \<and> nineteen \<noteq>
seventeen \<and> nineteen \<noteq> eighteen \<and> nineteen \<noteq>
twenty \<and> nineteen \<noteq> twentyone \<and> nineteen \<noteq>
twentytwo \<and> nineteen \<noteq> twentythree \<and> nineteen
\<noteq> twentyfour \<and> nineteen \<noteq> twentyfive \<and>
nineteen \<noteq> twentysix \<and> nineteen \<noteq> twentyseven
\<and> nineteen \<noteq> twentyeight \<and> nineteen \<noteq>
twentynine \<and> nineteen \<noteq> thirty \<and> nineteen \<noteq>
thirtyone \<and> twenty \<noteq> one \<and> twenty \<noteq> two \<and>
twenty \<noteq> three \<and> twenty \<noteq> four \<and> twenty
\<noteq> five \<and> twenty \<noteq> six \<and> twenty \<noteq> seven
\<and> twenty \<noteq> eight \<and> twenty \<noteq> nine \<and> twenty
\<noteq> ten \<and> twenty \<noteq> eleven \<and> twenty \<noteq>
twelve \<and> twenty \<noteq> thirteen \<and> twenty \<noteq> fourteen
\<and> twenty \<noteq> fifteen \<and> twenty \<noteq> sixteen \<and> twenty
\<noteq> seventeen \<and> twenty \<noteq> eighteen \<and> twenty
\<noteq> nineteen \<and> twenty \<noteq> twentyone \<and> twenty
\<noteq> twentytwo \<and> twenty \<noteq> twentythree \<and> twenty
\<noteq> twentyfour \<and> twenty \<noteq> twentyfive \<and> twenty
\<noteq> twentysix \<and> twenty \<noteq> twentyseven \<and> twenty
\<noteq> twentyeight \<and> twenty \<noteq> twentynine \<and> twenty
\<noteq> thirty \<and> twenty \<noteq> thirtyone \<and> twentyone
\<noteq> one \<and> twentyone \<noteq> two \<and> twentyone \<noteq>
three \<and> twentyone \<noteq> four \<and> twentyone \<noteq> five
\<and> twentyone \<noteq> six \<and> twentyone \<noteq> seven \<and>
twentyone \<noteq> eight \<and> twentyone \<noteq> nine \<and>
twentyone \<noteq> ten \<and> twentyone \<noteq> eleven \<and>
twentyone \<noteq> twelve \<and> twentyone \<noteq> thirteen \<and>
twentyone \<noteq> fourteen \<and> twentyone \<noteq> fifteen \<and>
twentyone \<noteq> sixteen \<and> twentyone \<noteq> seventeen \<and>
twentyone \<noteq> eighteen \<and> twentyone \<noteq> nineteen \<and>
twentyone \<noteq> twenty \<and> twentyone \<noteq> twentytwo \<and>
twentyone \<noteq> twentythree \<and> twentyone \<noteq> twentyfour
\<and> twentyone \<noteq> twentyfive \<and> twentyone \<noteq>
twentysix \<and> twentyone \<noteq> twentyseven \<and> twentyone
\<noteq> twentyeight \<and> twentyone \<noteq> twentynine \<and>
twentyone \<noteq> thirty \<and> twentyone \<noteq> thirtyone \<and>
twentytwo \<noteq> one \<and> twentytwo \<noteq> two \<and> twentytwo
\<noteq> three \<and> twentytwo \<noteq> four \<and> twentytwo
\<noteq> five \<and> twentytwo \<noteq> six \<and> twentytwo \<noteq>
seven \<and> twentytwo \<noteq> eight \<and> twentytwo \<noteq> nine
\<and> twentytwo \<noteq> ten \<and> twentytwo \<noteq> eleven \<and>
twentytwo \<noteq> twelve \<and> twentytwo \<noteq> thirteen \<and>
twentytwo \<noteq> fourteen \<and> twentytwo \<noteq> fifteen \<and>
twentytwo \<noteq> sixteen \<and> twentytwo \<noteq> seventeen \<and>
twentytwo \<noteq> eighteen \<and> twentytwo \<noteq> nineteen \<and>
twentytwo \<noteq> twenty \<and> twentytwo \<noteq> twentyone \<and>
twentytwo \<noteq> twentythree \<and> twentytwo \<noteq> twentyfour
\<and> twentytwo \<noteq> twentyfive \<and> twentytwo \<noteq>
twentysix \<and> twentytwo \<noteq> twentyseven \<and> twentytwo
\<noteq> twentyeight \<and> twentytwo \<noteq> twentynine \<and>
twentytwo \<noteq> thirty \<and> twentytwo \<noteq> thirtyone \<and>
twentythree \<noteq> one \<and> twentythree \<noteq> two \<and>
twentythree \<noteq> three \<and> twentythree \<noteq> four \<and>
twentythree \<noteq> five \<and> twentythree \<noteq> six \<and>
twentythree \<noteq> seven \<and> twentythree \<noteq> eight \<and>
twentythree \<noteq> nine \<and> twentythree \<noteq> ten \<and>
twentythree \<noteq> eleven \<and> twentythree \<noteq> twelve \<and>
twentythree \<noteq> thirteen \<and> twentythree \<noteq> fourteen \<and>
twentythree \<noteq> fifteen \<and> twentythree \<noteq> sixteen \<and>
twentythree \<noteq> seventeen \<and> twentythree \<noteq> eighteen
\<and> twentythree \<noteq> nineteen \<and> twentythree \<noteq>
twenty \<and> twentythree \<noteq> twentyone \<and> twentythree
\<noteq> twentytwo \<and> twentythree \<noteq> twentyfour \<and>
twentythree \<noteq> twentyfive \<and> twentythree \<noteq> twentysix
\<and> twentythree \<noteq> twentyseven \<and> twentythree \<noteq>
twentyeight \<and> twentythree \<noteq> twentynine \<and> twentythree
\<noteq> thirty \<and> twentythree \<noteq> thirtyone \<and>
twentyfour \<noteq> one \<and> twentyfour \<noteq> two \<and>
twentyfour \<noteq> three \<and> twentyfour \<noteq> four \<and>
twentyfour \<noteq> five \<and> twentyfour \<noteq> six \<and>
twentyfour \<noteq> seven \<and> twentyfour \<noteq> eight \<and>
twentyfour \<noteq> nine \<and> twentyfour \<noteq> ten \<and>
twentyfour \<noteq> eleven \<and> twentyfour \<noteq> twelve \<and>
twentyfour \<noteq> thirteen \<and> twentyfour \<noteq> fourteen \<and>
twentyfour \<noteq> fifteen \<and> twentyfour \<noteq> sixteen \<and>
twentyfour \<noteq> seventeen \<and> twentyfour \<noteq> eighteen
\<and> twentyfour \<noteq> nineteen \<and> twentyfour \<noteq> twenty
\<and> twentyfour \<noteq> twentyone \<and> twentyfour \<noteq>
twentytwo \<and> twentyfour \<noteq> twentythree \<and> twentyfour
\<noteq> twentyfive \<and> twentyfour \<noteq> twentysix \<and>
twentyfour \<noteq> twentyseven \<and> twentyfour \<noteq> twentyeight
\<and> twentyfour \<noteq> twentynine \<and> twentyfour \<noteq>
thirty \<and> twentyfour \<noteq> thirtyone \<and> twentyfive \<noteq>
one \<and> twentyfive \<noteq> two \<and> twentyfive \<noteq> three
\<and> twentyfive \<noteq> four \<and> twentyfive \<noteq> five \<and>
twentyfive \<noteq> six \<and> twentyfive \<noteq> seven \<and>
twentyfive \<noteq> eight \<and> twentyfive \<noteq> nine \<and>
twentyfive \<noteq> ten \<and> twentyfive \<noteq> eleven \<and>
twentyfive \<noteq> twelve \<and> twentyfive \<noteq> thirteen \<and>
twentyfive \<noteq> fourteen \<and> twentyfive \<noteq> fifteen \<and>
twentyfive \<noteq> sixteen \<and> twentyfive \<noteq> seventeen
\<and> twentyfive \<noteq> eighteen \<and> twentyfive \<noteq>
nineteen \<and> twentyfive \<noteq> twenty \<and> twentyfive \<noteq>
twentyone \<and> twentyfive \<noteq> twentytwo \<and> twentyfive
\<noteq> twentythree \<and> twentyfive \<noteq> twentyfour \<and>
twentyfive \<noteq> twentysix \<and> twentyfive \<noteq> twentyseven
\<and> twentyfive \<noteq> twentyeight \<and> twentyfive \<noteq>
twentynine \<and> twentyfive \<noteq> thirty \<and> twentyfive
\<noteq> thirtyone \<and> twentysix \<noteq> one \<and> twentysix
\<noteq> two \<and> twentysix \<noteq> three \<and> twentysix \<noteq>
four \<and> twentysix \<noteq> five \<and> twentysix \<noteq> six
\<and> twentysix \<noteq> seven \<and> twentysix \<noteq> eight \<and>
twentysix \<noteq> nine \<and> twentysix \<noteq> ten \<and> twentysix
\<noteq> eleven \<and> twentysix \<noteq> twelve \<and> twentysix
\<noteq> thirteen \<and> twentysix \<noteq> fourteen \<and> twentysix
\<noteq> fifteen \<and> twentysix \<noteq> sixteen \<and> twentysix
\<noteq> seventeen \<and> twentysix \<noteq> eighteen \<and> twentysix
\<noteq> nineteen \<and> twentysix \<noteq> twenty \<and> twentysix
\<noteq> twentyone \<and> twentysix \<noteq> twentytwo \<and>
twentysix \<noteq> twentythree \<and> twentysix \<noteq> twentyfour
\<and> twentysix \<noteq> twentyfive \<and> twentysix \<noteq>
twentyseven \<and> twentysix \<noteq> twentyeight \<and> twentysix
\<noteq> twentynine \<and> twentysix \<noteq> thirty \<and> twentysix
\<noteq> thirtyone \<and> twentyseven \<noteq> one \<and> twentyseven
\<noteq> two \<and> twentyseven \<noteq> three \<and> twentyseven
\<noteq> four \<and> twentyseven \<noteq> five \<and> twentyseven
\<noteq> six \<and> twentyseven \<noteq> seven \<and> twentyseven
\<noteq> eight \<and> twentyseven \<noteq> nine \<and> twentyseven
\<noteq> ten \<and> twentyseven \<noteq> eleven \<and> twentyseven
\<noteq> twelve \<and> twentyseven \<noteq> thirteen \<and> twentyseven
\<noteq> fourteen \<and> twentyseven \<noteq> fifteen \<and> twentyseven
\<noteq> sixteen \<and> twentyseven \<noteq> seventeen \<and>
twentyseven \<noteq> eighteen \<and> twentyseven \<noteq> nineteen
\<and> twentyseven \<noteq> twenty \<and> twentyseven \<noteq>
twentyone \<and> twentyseven \<noteq> twentytwo \<and> twentyseven
\<noteq> twentythree \<and> twentyseven \<noteq> twentyfour \<and>
twentyseven \<noteq> twentyfive \<and> twentyseven \<noteq> twentysix
\<and> twentyseven \<noteq> twentyeight \<and> twentyseven \<noteq>
twentynine \<and> twentyseven \<noteq> thirty \<and> twentyseven
\<noteq> thirtyone \<and> twentyeight \<noteq> one \<and> twentyeight
\<noteq> two \<and> twentyeight \<noteq> three \<and> twentyeight
\<noteq> four \<and> twentyeight \<noteq> five \<and> twentyeight
\<noteq> six \<and> twentyeight \<noteq> seven \<and> twentyeight
\<noteq> eight \<and> twentyeight \<noteq> nine \<and> twentyeight
\<noteq> ten \<and> twentyeight \<noteq> eleven \<and> twentyeight
\<noteq> twelve \<and> twentyeight \<noteq> thirteen \<and> twentyeight
\<noteq> fourteen \<and> twentyeight \<noteq> fifteen \<and> twentyeight
\<noteq> sixteen \<and> twentyeight \<noteq> seventeen \<and>
twentyeight \<noteq> eighteen \<and> twentyeight \<noteq> nineteen
\<and> twentyeight \<noteq> twenty \<and> twentyeight \<noteq>
twentyone \<and> twentyeight \<noteq> twentytwo \<and> twentyeight
\<noteq> twentythree \<and> twentyeight \<noteq> twentyfour \<and>
twentyeight \<noteq> twentyfive \<and> twentyeight \<noteq> twentysix
\<and> twentyeight \<noteq> twentyseven \<and> twentyeight \<noteq>
twentynine \<and> twentyeight \<noteq> thirty \<and> twentyeight
\<noteq> thirtyone \<and> twentynine \<noteq> one \<and> twentynine
\<noteq> two \<and> twentynine \<noteq> three \<and> twentynine
\<noteq> four \<and> twentynine \<noteq> five \<and> twentynine
\<noteq> six \<and> twentynine \<noteq> seven \<and> twentynine
\<noteq> eight \<and> twentynine \<noteq> nine \<and> twentynine
\<noteq> ten \<and> twentynine \<noteq> eleven \<and> twentynine
\<noteq> twelve \<and> twentynine \<noteq> thirteen \<and> twentynine
\<noteq> fourteen \<and> twentynine \<noteq> fifteen \<and> twentynine
\<noteq> sixteen \<and> twentynine \<noteq> seventeen \<and>
twentynine \<noteq> eighteen \<and> twentynine \<noteq> nineteen
\<and> twentynine \<noteq> twenty \<and> twentynine \<noteq> twentyone
\<and> twentynine \<noteq> twentytwo \<and> twentynine \<noteq>
twentythree \<and> twentynine \<noteq> twentyfour \<and> twentynine
\<noteq> twentyfive \<and> twentynine \<noteq> twentysix \<and>
twentynine \<noteq> twentyseven \<and> twentynine \<noteq> twentyeight
\<and> twentynine \<noteq> thirty \<and> twentynine \<noteq> thirtyone
\<and> thirty \<noteq> one \<and> thirty \<noteq> two \<and> thirty
\<noteq> three \<and> thirty \<noteq> four \<and> thirty \<noteq> five
\<and> thirty \<noteq> six \<and> thirty \<noteq> seven \<and> thirty
\<noteq> eight \<and> thirty \<noteq> nine \<and> thirty \<noteq> ten
\<and> thirty \<noteq> eleven \<and> thirty \<noteq> twelve \<and>
thirty \<noteq> thirteen \<and> thirty \<noteq> fourteen \<and> thirty
\<noteq> fifteen \<and> thirty \<noteq> sixteen \<and> thirty \<noteq>
seventeen \<and> thirty \<noteq> eighteen \<and> thirty \<noteq>
nineteen \<and> thirty \<noteq> twenty \<and> thirty \<noteq>
twentyone \<and> thirty \<noteq> twentytwo \<and> thirty \<noteq>
twentythree \<and> thirty \<noteq> twentyfour \<and> thirty \<noteq>
twentyfive \<and> thirty \<noteq> twentysix \<and> thirty \<noteq>
twentyseven \<and> thirty \<noteq> twentyeight \<and> thirty \<noteq>
twentynine \<and> thirty \<noteq> thirtyone \<and> thirtyone \<noteq>
one \<and> thirtyone \<noteq> two \<and> thirtyone \<noteq> three
\<and> thirtyone \<noteq> four \<and> thirtyone \<noteq> five \<and>
thirtyone \<noteq> six \<and> thirtyone \<noteq> seven \<and>
thirtyone \<noteq> eight \<and> thirtyone \<noteq> nine \<and>
thirtyone \<noteq> ten \<and> thirtyone \<noteq> eleven \<and>
thirtyone \<noteq> twelve \<and> thirtyone \<noteq> thirteen \<and>
thirtyone \<noteq> fourteen \<and> thirtyone \<noteq> fifteen \<and>
thirtyone \<noteq> sixteen \<and> thirtyone \<noteq> seventeen \<and>
thirtyone \<noteq> eighteen \<and> thirtyone \<noteq> nineteen \<and>
thirtyone \<noteq> twenty \<and> thirtyone \<noteq> twentyone \<and>
thirtyone \<noteq> twentytwo \<and> thirtyone \<noteq> twentythree
\<and> thirtyone \<noteq> twentyfour \<and> thirtyone \<noteq>
twentyfive \<and> thirtyone \<noteq> twentysix \<and> thirtyone
\<noteq> twentyseven \<and> thirtyone \<noteq> twentyeight \<and>
thirtyone \<noteq> twentynine \<and> thirtyone \<noteq> thirty"
apply (simp add: UnfoldPolicy) 
apply (auto simp: sets_distinct1) 
done
 

lemma aux: "\<lbrakk>x \<noteq> a; y\<noteq>b; (x \<noteq> y \<and> x \<noteq> b) \<or>  (a \<noteq> b \<and> a \<noteq> y)\<rbrakk> \<Longrightarrow> {x,a} \<noteq> {y,b}"
  apply auto
done


lemma aux2: "{a,b} = {b,a}"
  apply auto
done

 
consts fixID :: id
consts fixContent :: DummyContent

definition "fixElements p = (id p = fixID)"

lemmas fixDefs = fixElements_def NetworkCore.id_def NetworkCore.content_def



fun numberOfRules where 
 "numberOfRules (a\<oplus>b) = numberOfRules a + numberOfRules b"
 |"numberOfRules a = (1::int)" 


fun numberOfRulesList where 
 "numberOfRulesList  (x#xs) = ((numberOfRules x)#(numberOfRulesList xs)) "
 |"numberOfRulesList [] = []" 





text {* 
An example of how to test using the distributivity of the normalisation.
*}



definition Filter :: "(IntegerPort net,port) Combinators" where
"Filter = 
DenyAll \<oplus>  AllowPortFromTo one two 199 \<oplus> AllowPortFromTo one two 38 \<oplus> AllowPortFromTo one three 533 \<oplus> DenyAllFromTo two one\<oplus> AllowPortFromTo three two 578 \<oplus> AllowPortFromTo two one 645 \<oplus> AllowPortFromTo one three 694 \<oplus> AllowPortFromTo three one 871
"


lemma "dupl (policy2list Filter) = X"
apply (simp add: Filter_def)
oops



lemmas policies =  Filter_def 

lemmas UnfoldNetworkAndPolicy = UnfoldPolicy PLemmas policies


lemmas normalizeUnfold = makeFUT_def normalize_def policies Nets_List_def bothNets_def aux aux2 bothNets_def

lemmas casesSimp = fixDefs packet_Nets_def PLemmas UnfoldNetworkAndPolicy


lemma noMT2: "\<forall> x \<in> set (policy2list Filter). dom (C x) \<noteq> {}"
apply (simp add: UnfoldNetworkAndPolicy)
done


lemma count_the_rules: "(int (length(policy2list (list2policy(normalize Filter)))) = post) \<and> (int(length (policy2list Filter)) = pre) \<and>
 (int (length((normalize Filter))) = Partitions)"
apply (insert nets_different noMT2)
apply (simp add:  normalizeUnfold)
apply (thin_tac "?X")+
oops


lemma normalize: "normalize Filter = X"
apply (insert nets_different noMT2)
apply (simp add: normalizeUnfold, thin_tac "?X",thin_tac "?S")
oops


declare [[testgen_profiling]]



definition pat2pool_int :: "int set \<Rightarrow> int set \<Rightarrow> (IntegerPort,'\<beta>) packet \<mapsto> 
                                            (IntegerPort,'\<beta>) packet set" where 
 "pat2pool_int srcs transl = (\<lambda> p. if (fst (src p)) \<in> srcs then
                          (Some (allow {(i,(s1,s2),d,da). 
        (i = id p \<and> s1 \<in> transl \<and> d = dest p \<and> da = content p)}))
                              else None)" 



definition NAT1  ::"(IntegerPort,DummyContent) packet \<mapsto> 
     (((IntegerPort,DummyContent) packet) set)" where
 "NAT1 = pat2pool_int {7} {17,27,37}"


definition NAT2  ::"(IntegerPort,DummyContent) packet \<mapsto> 
          (((IntegerPort,DummyContent) packet) set)" where
 "NAT2 = pat2pool_int {6} {16,26,36}"

definition NATID :: "(IntegerPort,DummyContent) packet \<mapsto> 
        (((IntegerPort,DummyContent) packet) set)" where
 "NATID = (\<lambda>p. Some (allow {p}))"

definition NAT :: "(IntegerPort,DummyContent) packet \<mapsto> 
        (((IntegerPort,DummyContent) packet) set)" where
 "NAT = (NAT1 \<Oplus> NAT2 \<Oplus> NATID)"


definition Policy ::"(IntegerPort,DummyContent) packet \<mapsto> 
        (((IntegerPort,DummyContent) packet) set)" where
 "Policy = ((\<lambda> (x,y). x) o_f (NAT \<Otimes>\<^sub>2\<^sub>I (C Filter)))"



lemmas NAT = NAT2_def NATID_def NAT1_def



 text{*
Two auxiliary lemmas
*}
 

lemma all_in_list: "all_in_list (policy2list Filter) (Nets_List Filter)"
apply (simp add: Filter_def)
apply (unfold Nets_List_def)
apply (unfold bothNets_def)
apply (rule conjI)(*
apply (insert nets_different)
apply simp
apply (simp add: aux2)*)
sorry


lemma N3: "allNetsDistinct (policy2list Filter)"
sorry


lemma C_Filter_compile: " 
C (list2policy (normalize Filter)) =  C Filter"
apply (unfold normalize_def)
apply (rule C_eq_compile)
apply (simp add: Filter_def)
apply (rule all_in_list)
apply (rule N3)
done


lemma N1:  "dom NAT = UNIV"
apply (simp add: dom_def NAT NAT_def)
done


lemma N2: "DenyAll \<in> set (policy2list Filter)"
apply (simp add: Filter_def)
done


lemma N4: "all_in_list (policy2list Filter) (Nets_List Filter)"
apply (rule all_in_list)
done

 
lemma normFilter: "normalize Filter = [DenyAll,
     DenyAllFromTo three two \<oplus>
     DenyAllFromTo two three \<oplus> AllowPortFromTo three two 578,
     DenyAllFromTo one two \<oplus>
     ((DenyAllFromTo two one \<oplus> AllowPortFromTo one two 199) \<oplus>
      AllowPortFromTo one two 38) \<oplus>
     AllowPortFromTo two one 645,
     DenyAllFromTo one three \<oplus>
     DenyAllFromTo three one \<oplus>
     (AllowPortFromTo one three 533 \<oplus> AllowPortFromTo one three 694) \<oplus>
     AllowPortFromTo three one 871]"
(*apply (insert nets_different )
apply (simp add:  normalize_def Filter_def bothNets_def aux aux2 Nets_List_def)*)
sorry


lemma PolUniv: "dom Policy = UNIV"
apply (simp add: Policy_def)
apply auto
sorry 
(* Is true and easy to prove. Just not done yet...*)


lemma disjDomP: "disjDomGD ((NAT \<Otimes>\<^sub>2\<^sub>I\<^sub>L (Comb2PolList (rev (normalize (Filter))))))"
apply (simp add: normFilter)
apply (simp add: disjDomGD_def)
apply auto
sorry (* Is true but not proved yet. *)




test_spec "FUT x = Policy x"
apply (prepare_fw_spec)
apply (rule_tac PL = "NAT \<Otimes>\<^sub>2\<^sub>I\<^sub>L (Comb2PolList (rev (normalize (Filter))))" 
   in distrFUTLG1)
apply (simp add: normFilter)
apply (simp add: UNIV_I PolUniv)
apply (rule fun_cong) back
apply (simp add: Policy_def)
apply (rule normalizeNAT[symmetric])
apply (rule N1, rule N2, rule N3, rule N4)
apply (rule disjDomP)
apply (simp add: gatherDomain_def normFilter FUTListGD_def)
apply (rule conjI) defer 1
apply (rule conjI) defer 1
apply (rule conjI) 
(* Do some unfolding here. *)
apply (simp_all add: UnfoldPolicy fixDefs policies 
      UnfoldNetworkAndPolicy normalizeUnfold casesSimp 
      NAT PLemmas intranet_def internet_def C.simps 
      map_add_def dmz_def split_tupled_all
      prod_2_id_def prod_2_def NAT NAT_def pat2pool_int_def  
      policy_range_comp_def id_def
      content_def)(*
apply (gen_test_cases_old FUT)  (* 338 seconds *) (* 2158 seconds *)*)
store_test_thm "foos" 


test_spec " \<And>a aa b ab ba bb.
       (aa = 2 \<longrightarrow> ab \<noteq> 3) \<longrightarrow>
       (aa = 3 \<longrightarrow> ab \<noteq> 2) \<longrightarrow>
       (aa = 2 \<longrightarrow> ab \<noteq> 1) \<longrightarrow>
       (aa = 1 \<longrightarrow> ab \<noteq> 3) \<longrightarrow>
       (ba = 533 \<longrightarrow>
        (aa = 3 \<longrightarrow> ab \<noteq> 1) \<longrightarrow>
        (aa = 1 \<longrightarrow> ab \<noteq> 2) \<longrightarrow>
        (aa = 6 \<longrightarrow>
         FUT (a, (6, b), (ab, 533), bb) =
         Some (deny {(i, (s1, s2), aba).
                     i = a \<and>
                     (s1 = 16 \<or> s1 = 26 \<or> s1 = 36) \<and>
                     (case aba of (d, da) \<Rightarrow> d = (ab, 533) \<and> da = bb)})) \<and>
        (aa \<noteq> 6 \<longrightarrow>
         (aa = 7 \<longrightarrow>
          FUT (a, (7, b), (ab, 533), bb) =
          Some (deny {(i, (s1, s2), aba).
                      i = a \<and>
                      (s1 = 17 \<or> s1 = 27 \<or> s1 = 37) \<and>
                      (case aba of (d, da) \<Rightarrow> d = (ab, 533) \<and> da = bb)})) \<and>
         (aa \<noteq> 7 \<longrightarrow>
          FUT (a, (aa, b), (ab, 533), bb) =
          Some (deny {(a, (aa, b), (ab, 533), bb)})))) \<and>
       (ba \<noteq> 533 \<longrightarrow>
        (ba = 694 \<longrightarrow>
         (aa = 3 \<longrightarrow> ab \<noteq> 1) \<longrightarrow>
         (aa = 1 \<longrightarrow> ab \<noteq> 2) \<longrightarrow>
         (aa = 6 \<longrightarrow>
          FUT (a, (6, b), (ab, 694), bb) =
          Some (deny {(i, (s1, s2), aba).
                      i = a \<and>
                      (s1 = 16 \<or> s1 = 26 \<or> s1 = 36) \<and>
                      (case aba of (d, da) \<Rightarrow> d = (ab, 694) \<and> da = bb)})) \<and>
         (aa \<noteq> 6 \<longrightarrow>
          (aa = 7 \<longrightarrow>
           FUT (a, (7, b), (ab, 694), bb) =
           Some (deny {(i, (s1, s2), aba).
                       i = a \<and>
                       (s1 = 17 \<or> s1 = 27 \<or> s1 = 37) \<and>
                       (case aba of (d, da) \<Rightarrow> d = (ab, 694) \<and> da = bb)})) \<and>
          (aa \<noteq> 7 \<longrightarrow>
           FUT (a, (aa, b), (ab, 694), bb) =
           Some (deny {(a, (aa, b), (ab, 694), bb)})))) \<and>
        (ba \<noteq> 694 \<longrightarrow>
         (aa = 3 \<longrightarrow> ab \<noteq> 1) \<longrightarrow>
         (ba = 871 \<longrightarrow>
          (aa = 1 \<longrightarrow> ab \<noteq> 2) \<longrightarrow>
          (aa = 6 \<longrightarrow>
           FUT (a, (6, b), (ab, 871), bb) =
           Some (deny {(i, (s1, s2), aba).
                       i = a \<and>
                       (s1 = 16 \<or> s1 = 26 \<or> s1 = 36) \<and>
                       (case aba of (d, da) \<Rightarrow> d = (ab, 871) \<and> da = bb)})) \<and>
          (aa \<noteq> 6 \<longrightarrow>
           (aa = 7 \<longrightarrow>
            FUT (a, (7, b), (ab, 871), bb) =
            Some (deny {(i, (s1, s2), aba).
                        i = a \<and>
                        (s1 = 17 \<or> s1 = 27 \<or> s1 = 37) \<and>
                        (case aba of (d, da) \<Rightarrow> d = (ab, 871) \<and> da = bb)})) \<and>
           (aa \<noteq> 7 \<longrightarrow>
            FUT (a, (aa, b), (ab, 871), bb) =
            Some (deny {(a, (aa, b), (ab, 871), bb)})))) \<and>
         (ba \<noteq> 871 \<longrightarrow>
          (aa = 1 \<longrightarrow> ab \<noteq> 2) \<longrightarrow>
          (ba = 199 \<longrightarrow>
           (aa = 6 \<longrightarrow>
            FUT (a, (6, b), (ab, 199), bb) =
            Some (deny {(i, (s1, s2), aba).
                        i = a \<and>
                        (s1 = 16 \<or> s1 = 26 \<or> s1 = 36) \<and>
                        (case aba of (d, da) \<Rightarrow> d = (ab, 199) \<and> da = bb)})) \<and>
           (aa \<noteq> 6 \<longrightarrow>
            (aa = 7 \<longrightarrow>
             FUT (a, (7, b), (ab, 199), bb) =
             Some (deny {(i, (s1, s2), aba).
                         i = a \<and>
                         (s1 = 17 \<or> s1 = 27 \<or> s1 = 37) \<and>
                         (case aba of (d, da) \<Rightarrow> d = (ab, 199) \<and> da = bb)})) \<and>
            (aa \<noteq> 7 \<longrightarrow>
             FUT (a, (aa, b), (ab, 199), bb) =
             Some (deny {(a, (aa, b), (ab, 199), bb)})))) \<and>
          (ba \<noteq> 199 \<longrightarrow>
           (ba = 38 \<longrightarrow>
            (aa = 6 \<longrightarrow>
             FUT (a, (6, b), (ab, 38), bb) =
             Some (deny {(i, (s1, s2), aba).
                         i = a \<and>
                         (s1 = 16 \<or> s1 = 26 \<or> s1 = 36) \<and>
                         (case aba of (d, da) \<Rightarrow> d = (ab, 38) \<and> da = bb)})) \<and>
            (aa \<noteq> 6 \<longrightarrow>
             (aa = 7 \<longrightarrow>
              FUT (a, (7, b), (ab, 38), bb) =
              Some (deny {(i, (s1, s2), aba).
                          i = a \<and>
                          (s1 = 17 \<or> s1 = 27 \<or> s1 = 37) \<and>
                          (case aba of (d, da) \<Rightarrow> d = (ab, 38) \<and> da = bb)})) \<and>
             (aa \<noteq> 7 \<longrightarrow>
              FUT (a, (aa, b), (ab, 38), bb) =
              Some (deny {(a, (aa, b), (ab, 38), bb)})))) \<and>
           (ba \<noteq> 38 \<longrightarrow>
            (ba = 645 \<longrightarrow>
             (aa = 6 \<longrightarrow>
              FUT (a, (6, b), (ab, 645), bb) =
              Some (deny {(i, (s1, s2), aba).
                          i = a \<and>
                          (s1 = 16 \<or> s1 = 26 \<or> s1 = 36) \<and>
                          (case aba of (d, da) \<Rightarrow> d = (ab, 645) \<and> da = bb)})) \<and>
             (aa \<noteq> 6 \<longrightarrow>
              (aa = 7 \<longrightarrow>
               FUT (a, (7, b), (ab, 645), bb) =
               Some (deny {(i, (s1, s2), aba).
                           i = a \<and>
                           (s1 = 17 \<or> s1 = 27 \<or> s1 = 37) \<and>
                           (case aba of (d, da) \<Rightarrow> d = (ab, 645) \<and> da = bb)})) \<and>
              (aa \<noteq> 7 \<longrightarrow>
               FUT (a, (aa, b), (ab, 645), bb) =
               Some (deny {(a, (aa, b), (ab, 645), bb)})))) \<and>
            (ba \<noteq> 645 \<longrightarrow>
             (ba = 578 \<longrightarrow>
              (aa = 6 \<longrightarrow>
               FUT (a, (6, b), (ab, 578), bb) =
               Some (deny {(i, (s1, s2), aba).
                           i = a \<and>
                           (s1 = 16 \<or> s1 = 26 \<or> s1 = 36) \<and>
                           (case aba of (d, da) \<Rightarrow> d = (ab, 578) \<and> da = bb)})) \<and>
              (aa \<noteq> 6 \<longrightarrow>
               (aa = 7 \<longrightarrow>
                FUT (a, (7, b), (ab, 578), bb) =
                Some (deny {(i, (s1, s2), aba).
                            i = a \<and>
                            (s1 = 17 \<or> s1 = 27 \<or> s1 = 37) \<and>
                            (case aba of (d, da) \<Rightarrow> d = (ab, 578) \<and> da = bb)})) \<and>
               (aa \<noteq> 7 \<longrightarrow>
                FUT (a, (aa, b), (ab, 578), bb) =
                Some (deny {(a, (aa, b), (ab, 578), bb)})))) \<and>
             (ba \<noteq> 578 \<longrightarrow>
              (aa = 6 \<longrightarrow>
               FUT (a, (6, b), (ab, ba), bb) =
               Some (deny {(i, (s1, s2), aba).
                           i = a \<and>
                           (s1 = 16 \<or> s1 = 26 \<or> s1 = 36) \<and>
                           (case aba of (d, da) \<Rightarrow> d = (ab, ba) \<and> da = bb)})) \<and>
              (aa \<noteq> 6 \<longrightarrow>
               (aa = 7 \<longrightarrow>
                FUT (a, (7, b), (ab, ba), bb) =
                Some (deny {(i, (s1, s2), aba).
                            i = a \<and>
                            (s1 = 17 \<or> s1 = 27 \<or> s1 = 37) \<and>
                            (case aba of (d, da) \<Rightarrow> d = (ab, ba) \<and> da = bb)})) \<and>
               (aa \<noteq> 7 \<longrightarrow>
                FUT (a, (aa, b), (ab, ba), bb) =
                Some (deny {(a, (aa, b), (ab, ba), bb)}))))))))))"
apply (gen_test_cases FUT)
gen_test_data "foo"

thm foo.test_data


test_spec "FUT x = Policy x"
apply (prepare_fw_spec)
apply (simp_all add: UnfoldPolicy fixDefs policies  Policy_def
      UnfoldNetworkAndPolicy normalizeUnfold casesSimp 
      NAT PLemmas intranet_def internet_def C.simps 
      map_add_def dmz_def split_tupled_all
      prod_2_id_def prod_2_def NAT NAT_def pat2pool_int_def  
      policy_range_comp_def id_def
      content_def)
apply (gen_test_cases_old FUT)  
apply simp_all
store_test_thm "foo" 

gen_test_data "foo"

thm foo.test_data


write_clocks "time_stats_NATNormalisation"


test_spec "FUT x = Policy x"
apply (prepare_fw_spec)
(* Do some unfolding here. *)
apply (simp_all add: Policy_def Filter_def
      PLemmas intranet_def internet_def C.simps 
      map_add_def dmz_def split_tupled_all
      prod_2_id_def prod_2_def NAT NAT_def pat2pool_int_def  
      policy_range_comp_def id_def
      content_def)
apply (gen_test_cases_old FUT)  (* 338 seconds *)
apply simp_all
store_test_thm "without" 

gen_test_data "without"

thm without.test_data


end
